use std::{io, path::Path, str::Utf8Error};

use serde::{Deserialize, Serialize};
use serde_json::Value;
use thiserror::Error;
use tokio::{fs, process::Command};
use tracing::{error, info, warn};
use url::Url;
use utoipa::ToSchema;

use crate::string::ToUnquotedString;

#[derive(Debug, Error)]
pub enum ChannelError {
    #[error("URL is an invalid YouTube channel URL")]
    InvalidUrl,
    #[error("Channel already exists in database")]
    AlreadyExists,
    #[error("IO error: {0}")]
    IOError(#[from] io::Error),
    #[error("UTF-8 error: {0}")]
    Utf8Error(#[from] Utf8Error),
    #[error("Got erroneous output for yt-dlp task")]
    TaskFailed,
    #[error("Failed to extract channel ID from '{0}'")]
    ExtractID(String),
    #[error("Could not serialize info JSON: {0}")]
    SerializeInfoJSON(#[from] serde_json::Error),
    #[error("Failed to parse value from key '{0}'")]
    JsonKey(String),
}

#[derive(Serialize, Deserialize, Clone, ToSchema)]
pub struct Channel {
    pub id: i64,
    pub url: String,
    pub youtube_id: String,
    pub name: String,
    pub handle_url: String,
    pub avatar_url: String,
    pub banner_url: String,
    pub description: String,
    pub subscribers: i64,
}

impl Channel {
    async fn run_task(url: &str) -> Result<(), ChannelError> {
        let mut child = Command::new("yt-dlp")
            .args([
                "--write-info-json",
                "--skip-download",
                "--playlist-items",
                "1",
                "-v",
                "-o",
                "media/channels/tmp/tmp.%(ext)s",
                url,
            ])
            .spawn()?;
        info!("yt-dlp task invoked");
        child.wait().await.map_err(|_| ChannelError::TaskFailed)?;
        info!("yt-dlp task completed successfully");

        Ok(())
    }

    async fn run_update_task(url: &str) -> Result<(), ChannelError> {
        let mut child = Command::new("yt-dlp")
            .args([
                "--write-info-json",
                "--skip-download",
                "--playlist-items",
                "1",
                "-v",
                "-o",
                "media/channels/%(id)s/%(id)s.%(ext)s",
                url,
            ])
            .spawn()?;
        info!("yt-dlp task invoked");
        child.wait().await.map_err(|_| ChannelError::TaskFailed)?;
        info!("yt-dlp task completed successfully");

        Ok(())
    }

    pub async fn update_metadata(&mut self) -> Result<(), ChannelError> {
        Self::run_update_task(&self.url).await?;

        let json_path = format!("media/channels/{0}/{0}.info.json", self.youtube_id);
        let info: Value = serde_json::from_str(&fs::read_to_string(&json_path).await?)?;

        self.name = Self::get_json_value(&info, "channel")?.to_unquoted_string();
        self.handle_url = Self::get_json_value(&info, "uploader_url")?.to_unquoted_string();
        (self.avatar_url, self.banner_url) = Self::get_thumbnails(&info);
        self.description = Self::get_json_value(&info, "description")?.to_unquoted_string();
        self.subscribers = Self::get_json_value(&info, "channel_follower_count")?
            .as_i64()
            .unwrap_or_default();

        info!(
            "New information for '{}': {{ name: {}, handle_url: {}, avatar_url: {}, banner_url: {}, description: {}, subscribers: {} }}",
            self.youtube_id,
            self.name,
            self.handle_url,
            self.avatar_url,
            self.banner_url,
            self.description,
            self.subscribers
        );

        Ok(())
    }

    /// Transforms a channel handle or username into a canonical channel ID.
    ///
    /// While it invariably returns a canonical ID even if the URL already contains one,
    /// it's not recommended to call this function when that is the case in order to
    /// prevent multiple calls to yt-dlp.
    async fn extract_id(url: &str) -> Result<String, ChannelError> {
        let output = Command::new("yt-dlp")
            .args([
                "--skip-download",
                "--playlist-items",
                "1",
                "--print",
                "channel_id",
                url,
            ])
            .output()
            .await?;

        if !output.status.success() {
            error!(
                "yt-dlp could not download channel. Does the URL point to a valid/existing channel?"
            );
            return Err(ChannelError::TaskFailed);
        }

        let stdout = str::from_utf8(&output.stdout)?;
        let mut lines = stdout.lines().filter(|line| line.starts_with("UC"));

        lines.next().map_or_else(
            || Err(ChannelError::ExtractID(url.into())),
            |id| Ok(id.into()),
        )
    }

    fn get_json_value(json: &Value, key: &str) -> Result<Value, ChannelError> {
        json.get(key)
            .ok_or_else(|| ChannelError::JsonKey(key.to_string()))
            .cloned()
    }

    fn get_thumbnails(info_json: &Value) -> (String, String) {
        let mut avatar_url = String::new();
        let mut banner_url = String::new();

        if let Some(thumbnails) = info_json.get("thumbnails").and_then(Value::as_array) {
            for thumbnail in thumbnails {
                if let Some(id) = thumbnail.get("id").and_then(Value::as_str) {
                    match id {
                        "avatar_uncropped" => {
                            avatar_url = thumbnail
                                .get("url")
                                .and_then(Value::as_str)
                                .unwrap_or_default()
                                .to_string();
                        }
                        "banner_uncropped" => {
                            banner_url = thumbnail
                                .get("url")
                                .and_then(Value::as_str)
                                .unwrap_or_default()
                                .to_string();
                        }
                        _ => {}
                    }
                }
            }
        }

        (avatar_url, banner_url)
    }

    pub async fn from_url(url: &Url, id: i64) -> Result<Self, ChannelError> {
        let url_path = url.path();

        if !url_path.starts_with("/@")
            && !url_path.starts_with("/c/")
            && !url_path.starts_with("/channel/")
        {
            return Err(ChannelError::InvalidUrl);
        }

        let url = url.as_str();
        let youtube_id = if let Some(id) = url_path.strip_prefix("/channel/") {
            if id.starts_with("UC") {
                id.into()
            } else {
                warn!("Channel ID contains unexpected format: {id}, extracting");
                Self::extract_id(url).await?
            }
        } else {
            warn!("URL is not canonical, extracting ID");
            Self::extract_id(url).await?
        };

        let dir = format!("media/channels/{youtube_id}");
        let file_stem = format!("{dir}/{youtube_id}");

        if Path::new(&dir).exists() {
            warn!("Channel already exists, skipping");
            return Err(ChannelError::AlreadyExists);
        }

        fs::create_dir(dir).await?;

        Self::run_task(url).await?;
        let info: Value =
            serde_json::from_str(&fs::read_to_string("media/channels/tmp/tmp.info.json").await?)?;

        let url = Self::get_json_value(&info, "channel_url")?.to_unquoted_string();
        let name = Self::get_json_value(&info, "channel")?.to_unquoted_string();
        let handle_url = Self::get_json_value(&info, "uploader_url")?.to_unquoted_string();

        let (avatar_url, banner_url) = Self::get_thumbnails(&info);
        if avatar_url.is_empty() {
            warn!("Channel {youtube_id} has no avatar!");
        }
        if banner_url.is_empty() {
            warn!("Channel {youtube_id} has no banner!");
        }
        let description = Self::get_json_value(&info, "description")?.to_unquoted_string();
        let subscribers = Self::get_json_value(&info, "channel_follower_count")?
            .as_i64()
            .unwrap_or_default();

        fs::rename(
            "media/channels/tmp/tmp.info.json",
            format!("{file_stem}.info.json"),
        )
        .await?;

        Ok(Self {
            id,
            url,
            youtube_id,
            name,
            handle_url,
            avatar_url,
            banner_url,
            description,
            subscribers,
        })
    }
}
