use std::net::SocketAddr;

use axum::{
    Router,
    routing::{delete, get, post, put},
};
use doc::docs_router;
use instance::Instance;

use routes::{
    channel::{
        delete_channel, get_channel, get_channel_videos, list_channels, update_channel,
        upload_channel,
    },
    comment::get_video_comments,
    index,
    middleware::auth,
    video::{delete_video, get_video, list_videos, update_video, upload_video},
};
use tokio::signal;
use tracing::info;

mod channel;
mod comment;
mod config;
mod doc;
mod instance;
mod routes;
mod string;
mod url;
mod video;

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    tracing_subscriber::fmt::init();
    dotenvy::dotenv()?;

    let instance = Instance::new().await?;

    info!(
        "Initialized instance successfully!\nConfiguration:\n- Address\n  + Host: {}\n  + Port: {}\n- Pagination\n  + Videos per page: {}\n  + Comments per page: {}\n  + Channels per page: {}",
        instance.config.address.host,
        instance.config.address.port,
        instance.config.pagination.videos,
        instance.config.pagination.comments,
        instance.config.pagination.channels,
    );

    let address = instance.config.address.get_url();

    let router = Router::new()
        .route("/", get(index))
        .route("/video", get(list_videos))
        .route("/video/{id}", get(get_video))
        .route("/video/{id}/comments", get(get_video_comments))
        .route("/channel", get(list_channels))
        .route("/channel/{id}", get(get_channel))
        .route("/channel/{id}/videos", get(get_channel_videos));

    let protected = Router::new()
        .route("/video", post(upload_video))
        .route("/video/{id}", put(update_video))
        .route("/video/{id}", delete(delete_video))
        .route("/channel", post(upload_channel))
        .route("/channel/{id}", put(update_channel))
        .route("/channel/{id}", delete(delete_channel))
        .route_layer(axum::middleware::from_fn_with_state(instance.clone(), auth));

    let almond = router
        .merge(protected)
        .merge(docs_router())
        .with_state(instance);

    let listener = tokio::net::TcpListener::bind(address).await?;

    axum::serve(
        listener,
        almond.into_make_service_with_connect_info::<SocketAddr>(),
    )
    .with_graceful_shutdown(shutdown_signal())
    .await?;

    Ok(())
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("Failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        () = ctrl_c => {},
        () = terminate => {},
    }
}
