use axum::{
    Json,
    extract::{Path, Query, State},
    http::StatusCode,
};
use serde::{Deserialize, Serialize};
use tracing::{error, info};
use url::Url;
use utoipa::ToSchema;

use crate::{
    channel::{Channel, ChannelError},
    instance::Instance,
    url::is_youtube_url,
    video::Video,
};

#[derive(Deserialize, ToSchema)]
pub struct ListChannelsQuery {
    page: Option<usize>,
}

#[derive(Serialize, ToSchema)]
pub struct ListChannelsResponse {
    channels: Vec<Channel>,
    page: usize,
    per_page: usize,
    total: usize,
    pages: usize,
}

/// Retrieve video list as JSON (paged)
#[utoipa::path(get, path = "/channel", request_body = ListChannelsQuery, params(("page" = Option<usize>, Query)), tags = ["channel"], responses((status = 200, body = ListChannelsResponse), (status = 500, description = "Error fetching channels from database")))]
pub async fn list_channels(
    State(state): State<Instance>,
    Query(query): Query<ListChannelsQuery>,
) -> Result<Json<ListChannelsResponse>, StatusCode> {
    let Ok(channels) = sqlx::query_as!(Channel, "SELECT * FROM channel")
        .fetch_all(&state.pool)
        .await
    else {
        error!("Could not fetch channels from database!");
        return Err(StatusCode::INTERNAL_SERVER_ERROR);
    };

    let per_page = state.config.pagination.channels;
    let total = channels.len();
    let pages = total.div_ceil(per_page);
    let page = query.page.unwrap_or(1).min(pages).max(1);

    let start = per_page * (page - 1);
    let end = (start + per_page).min(total);

    let channels = if start < total {
        channels[start..end].to_vec()
    } else {
        vec![]
    };

    Ok(Json(ListChannelsResponse {
        channels,
        page,
        per_page,
        total,
        pages,
    }))
}

/// Get a single channel from the database by its ID
#[utoipa::path(get, path = "/channel/{id}", tags = ["channel"], responses((status = 200, body = Channel), (status = 404, description = "Not found"), (status = 500, description = "Failed to get channel from database")))]
pub async fn get_channel(
    State(state): State<Instance>,
    Path(id): Path<String>,
) -> Result<Json<Channel>, StatusCode> {
    sqlx::query_as!(Channel, "SELECT * FROM channel WHERE youtube_id = ?", id)
        .fetch_optional(&state.pool)
        .await
        .map_or(Err(StatusCode::INTERNAL_SERVER_ERROR), |channel| {
            channel.map_or(Err(StatusCode::NOT_FOUND), |c| Ok(Json(c)))
        })
}

/// Get all videos belonging to a channel ID from the database
#[utoipa::path(get, path = "/channel/{id}/videos", tags = ["channel"], responses((status = 200, body = Vec<Video>), (status = 404, description = "Not found"), (status = 500, description = "Error fetching videos from database")))]
pub async fn get_channel_videos(
    State(state): State<Instance>,
    Path(id): Path<String>,
) -> Result<Json<Vec<Video>>, StatusCode> {
    sqlx::query_as!(Video, "SELECT * FROM video WHERE author_id = ?", id)
        .fetch_all(&state.pool)
        .await
        .map_or(Err(StatusCode::INTERNAL_SERVER_ERROR), |videos| {
            if videos.is_empty() {
                Err(StatusCode::NOT_FOUND)
            } else {
                Ok(Json(videos))
            }
        })
}

#[derive(Deserialize, ToSchema)]
pub struct UploadChannelQuery {
    url: String,
}

/// Upload a channel's metadata to the database
#[utoipa::path(post, path = "/channel", request_body = UploadChannelQuery, params(("url" = String, Query)), tags = ["channel"], responses((status = 200, description = "Channel already exists, skipping"), (status = 201, description = "Channel uploaded successfully"), (status = 400, description = "Bad request, likely a malformed URL"), (status = 401, description = "Unauthorized"), (status = 500, description = "Failed to upload channel to database")), security(("authKey" = [])))]
pub async fn upload_channel(
    State(state): State<Instance>,
    Query(query): Query<UploadChannelQuery>,
) -> StatusCode {
    let id = match sqlx::query_scalar!("SELECT MAX(id) FROM channel")
        .fetch_one(&state.pool)
        .await
    {
        Ok(Some(max_id)) => max_id + 1,
        Ok(None) => 0,
        Err(_) => {
            return StatusCode::INTERNAL_SERVER_ERROR;
        }
    };

    let Ok(url) = Url::parse(&query.url) else {
        error!("Could not parse URL!");
        return StatusCode::BAD_REQUEST;
    };

    if !is_youtube_url(&url) {
        error!("YouTube URL RegEx match failed!");
        return StatusCode::BAD_REQUEST;
    }

    let new_channel = Channel::from_url(&url, id).await.map_err(|e| match e {
        ChannelError::AlreadyExists => StatusCode::OK,
        ChannelError::TaskFailed => StatusCode::BAD_REQUEST,
        e => {
            error!("{e:?}");
            StatusCode::INTERNAL_SERVER_ERROR
        }
    });

    match new_channel {
        Ok(channel) => {
            match sqlx::query!(
                "
                INSERT INTO channel (
                    id, url, youtube_id, name, handle_url, avatar_url, 
                    banner_url, description, subscribers
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                ",
                channel.id,
                channel.url,
                channel.youtube_id,
                channel.name,
                channel.handle_url,
                channel.avatar_url,
                channel.banner_url,
                channel.description,
                channel.subscribers,
            )
            .execute(&state.pool)
            .await
            {
                Ok(result) => {
                    info!("Inserted channel to database successfully! {result:?}");
                    StatusCode::CREATED
                }
                Err(e) => {
                    error!("Error inserting channel to database: {e:?}");
                    StatusCode::INTERNAL_SERVER_ERROR
                }
            }
        }
        Err(status) => status,
    }
}

/// Update an existing channel from the database
#[utoipa::path(put, path = "/channel/{id}", params(("almond-api-key" = String, description = "API key")), tags = ["channel"], responses((status = 204, description = "Channel updated successfully"), (status = 404, description = "Video with specified ID not found"), (status = 500, description = "Failed to update channel from database")), security(("authKey" = [])))]
pub async fn update_channel(State(state): State<Instance>, Path(id): Path<String>) -> StatusCode {
    match sqlx::query_as!(Channel, "SELECT * FROM channel WHERE youtube_id = ?", id)
        .fetch_optional(&state.pool)
        .await
        .map_or(Err(StatusCode::INTERNAL_SERVER_ERROR), |channel| {
            channel.ok_or(StatusCode::NOT_FOUND)
        }) {
        Ok(mut channel) => match channel.update_metadata().await {
            Ok(()) => {
                match sqlx::query!(
                    "UPDATE channel SET name = ?, handle_url = ?, avatar_url = ?, banner_url = ?, description = ?, subscribers = ? WHERE youtube_id = ?",
                    channel.name,
                    channel.handle_url,
                    channel.avatar_url,
                    channel.banner_url,
                    channel.description,
                    channel.subscribers,
                    id
                ).execute(&state.pool).await {
                    Ok(r) => {
                        info!("Updated channel '{id}' successfully: {r:?}");
                        StatusCode::NO_CONTENT
                    },
                    Err(e) => {
                        error!("Update channel failed: {e:?}");
                        StatusCode::INTERNAL_SERVER_ERROR
                    }
                }
            }
            Err(e) => {
                error!("Update video failed: {e:?}");
                StatusCode::INTERNAL_SERVER_ERROR
            }
        },
        Err(status) => {
            if status != StatusCode::NOT_FOUND {
                error!("Failed to update video '{id}'");
            }
            status
        }
    }
}

/// Delete a channel from the database
#[utoipa::path(delete, path = "/channel/{id}", params(("almond-api-key" = String, description = "API key")), tags = ["channel"], responses((status = 204, description = "Channel deleted successfully"), (status = 404, description = "Video with specified ID not found"), (status = 500, description = "Failed to delete channel from database")), security(("authKey" = [])))]
pub async fn delete_channel(State(state): State<Instance>, Path(id): Path<String>) -> StatusCode {
    match sqlx::query!("DELETE FROM channel WHERE youtube_id = ?", id)
        .fetch_optional(&state.pool)
        .await
        .map_or(Err(StatusCode::INTERNAL_SERVER_ERROR), |channel| {
            channel.ok_or(StatusCode::NOT_FOUND)
        }) {
        Ok(_) => {
            info!("Deleted channel '{id}' successfully");
            StatusCode::OK
        }
        Err(status) => status,
    }
}
