use std::{io, path::Path};

use serde::{Deserialize, Serialize};
use serde_json::Value;
use sha3::{Digest, Sha3_256};
use thiserror::Error;
use tokio::fs;
use tracing::{error, info, warn};
use url::Url;
use utoipa::ToSchema;

use crate::string::ToUnquotedString;

#[derive(Debug, Error)]
pub enum VideoError {
    #[error("URL is an invalid YouTube URL")]
    InvalidUrl,
    #[error("Video already exists in database")]
    AlreadyExists,
    #[error("IO Error: {0}")]
    IOError(#[from] io::Error),
    #[error("Got erroneous status code for yt-dlp task")]
    TaskFailed,
    #[error("Could not serialize info JSON: {0}")]
    SerializeInfoJSON(#[from] serde_json::Error),
    #[error("Failed to parse value from key '{0}'")]
    JsonKey(String),
    #[error("No video file was found for target")]
    MissingVideoFile,
}

#[derive(Serialize, Deserialize, Clone, ToSchema)]
pub struct Video {
    pub id: i64,
    pub url: String,
    pub youtube_id: String,
    pub title: String,
    pub description: String,
    pub author: String,
    pub author_id: String,
    pub author_url: String,
    pub views: i64,
    pub upload_date: String,
    pub likes: Option<i64>,
    pub dislikes: Option<i64>,
    pub file_name: String,
    pub file_size: i64,
    pub sha256: String,
    pub thumbnail: String,
    // pub comments: Vec<Comment>
}

impl Video {
    async fn run_task(url: &str) -> Result<(), VideoError> {
        let args = vec![
            "--write-info-json",
            "--write-thumbnail",
            "--write-description",
            "--write-comments",
            "--no-playlist",
            "--use-postprocessor",
            "ReturnYoutubeDislike:when=pre_process",
            "-v",
            url,
            "-o",
            "media/videos/%(id)s/%(id)s.%(ext)s",
            "-f",
            "bestvideo[ext=mkv]+bestaudio[ext=m4a]/bestvideo[ext=mp4]+bestaudio[ext=m4a]/bestvideo*+bestaudio/best",
        ];
        let mut child = tokio::process::Command::new("yt-dlp").args(args).spawn()?;
        info!("yt-dlp task invoked");
        child.wait().await.map_err(|_| VideoError::TaskFailed)?;
        info!("yt-dlp task completed successfully");

        Ok(())
    }

    async fn run_update_task(url: &str) -> Result<(), VideoError> {
        let args = vec![
            "--write-info-json",
            "--skip-download",
            "--write-thumbnail",
            "--write-description",
            "--write-comments",
            "--force-write",
            "--no-playlist",
            "--use-postprocessor",
            "ReturnYoutubeDislike:when=pre_process",
            "-v",
            url,
            "-o",
            "media/videos/%(id)s/%(id)s.%(ext)s",
        ];
        let mut child = tokio::process::Command::new("yt-dlp").args(args).spawn()?;
        info!("yt-dlp task invoked");
        child.wait().await.map_err(|_| VideoError::TaskFailed)?;
        info!("yt-dlp task completed successfully");

        Ok(())
    }

    pub async fn update_metadata(&mut self) -> Result<(), VideoError> {
        Self::run_update_task(&self.url).await?;

        let json_path = format!("media/videos/{0}/{0}.info.json", self.youtube_id);

        let info: Value = serde_json::from_str(&fs::read_to_string(&json_path).await?)?;

        self.title = Self::get_json_value(&info, "title")?.to_unquoted_string();
        self.views = Self::get_json_value(&info, "view_count")?
            .as_i64()
            .unwrap_or(-1);

        // Use RYD field from info JSON
        self.likes = Self::get_json_value(&info["RYD"]["response"], "likes")?.as_i64();
        self.dislikes = Self::get_json_value(&info["RYD"]["response"], "dislikes")?.as_i64();

        info!(
            "New information for video '{}': {{ title: {}, views: {}, likes: {:?}, dislikes: {:?} }}",
            self.youtube_id, self.title, self.views, self.likes, self.dislikes
        );

        Ok(())
    }

    fn get_json_value(json: &Value, key: &str) -> Result<Value, VideoError> {
        json.get(key)
            .ok_or_else(|| VideoError::JsonKey(key.to_string()))
            .cloned()
    }

    pub async fn from_url(url: &Url, id: i64) -> Result<Self, VideoError> {
        let mut pairs = url.query_pairs();
        let Some(query_v) = pairs.find(|(key, _)| key == "v") else {
            error!("Could not find 'v' query parameter in URL!");
            return Err(VideoError::InvalidUrl);
        };

        let youtube_id = query_v.1.to_string();
        info!("URL is valid YouTube video, got ID '{youtube_id}'");

        let dir = format!("media/videos/{youtube_id}");
        let file_stem = format!("{dir}/{youtube_id}");

        // ? Uploading a video doesn't mean updating it, make a PUT route for that later
        if Path::new(&dir).exists() {
            warn!("Video already exists, skipping");
            return Err(VideoError::AlreadyExists);
        }

        fs::create_dir(dir).await?;

        let info_json = format!("{file_stem}.info.json");
        let info_json = Path::new(&info_json);

        Self::run_task(url.as_str()).await?;

        let info: Value = serde_json::from_str(&fs::read_to_string(info_json).await?)?;

        // info!("Info JSON for {youtube_id}\n{info}");

        let url = url.to_string();
        let description = format!("{file_stem}.description");
        let title = Self::get_json_value(&info, "title")?.to_unquoted_string();
        let author = Self::get_json_value(&info, "uploader")?.to_unquoted_string();
        let views = Self::get_json_value(&info, "view_count")?
            .as_i64()
            .unwrap_or(-1);

        // Use RYD field from info JSON
        let likes = Self::get_json_value(&info["RYD"]["response"], "likes")?.as_i64();
        let dislikes = Self::get_json_value(&info["RYD"]["response"], "dislikes")?.as_i64();

        let possible_extensions = ["mkv", "mp4", "webm"];

        let mut buffer = None;
        let mut file_name = String::new();

        for ext in possible_extensions {
            file_name = format!("{file_stem}.{ext}");
            let path = Path::new(&file_name);
            if path.exists() {
                info!("File '{}' found", path.display());
                buffer = Some(fs::read(path).await?);
                break;
            }
            warn!("File '{}' not found", path.display());
        }

        let Some(buffer) = buffer else {
            error!("No video file found for {youtube_id}!");
            return Err(VideoError::MissingVideoFile);
        };

        let sha256 = format!("{:x}", Sha3_256::digest(&buffer));
        #[allow(clippy::cast_possible_wrap)]
        let file_size = buffer.len() as i64;
        let author_id = Self::get_json_value(&info, "channel_id")?.to_unquoted_string();
        let author_url = Self::get_json_value(&info, "channel_url")?.to_unquoted_string();
        let upload_date = Self::get_json_value(&info, "upload_date")?.to_unquoted_string();

        let thumbnail = format!("{file_stem}.webp");

        let video = Self {
            id,
            url,
            youtube_id,
            title,
            description,
            author,
            author_id,
            author_url,
            views,
            upload_date,
            likes,
            dislikes,
            file_name,
            file_size,
            sha256,
            thumbnail,
        };

        Ok(video)
    }
}
